/*
 * ARM Integrator CP System emulation.
 *
 * Copyright (c) 2005 CodeSourcery, LLC.
 * Written by Paul Brook
 *
 * This code is licenced under the GPL
 */
#include <assert.h>

#include "vl.h"
#include "arm_pic.h"

#include "i2c.h"
#include "vidc.h"

#define DEBUG_IO

#define MACHINE_RISCPC
//#define MACHINE_A7000
//#define MACHINE_A7000PLUS

#if defined(MACHINE_RISCPC)
# define CHIP_ID 0xd4e7
#elif defined(MACHINE_A7000)
# define CHIP_ID 0x5b98
#elif defined(MACHINE_A7000PLUS)
# define CHIP_ID 0xaa7c
#else
# error Unknown machine for Chip ID
#endif

#ifdef DEBUG_IO
# define DPRINTF(fmt, args...) \
  do { fprintf(stderr, fmt, ##args); } while (0)
#else
# define DPRINTF(fmt, args...)
#endif

extern void smc37c665_mm_init(target_ulong base, unsigned it_shift);

typedef struct {
    /* Programmable Interrupt Controller fields */
    arm_pic_handler handler;    /* Must be first element of struct */
    void *parent;
    int parent_irq;
    int parent_fiq;

    CPUState *env;

    uint8_t io_control;

    /* Keyboard registers */
    uint8_t kbd_control;

    /* IRQ A registers */
    uint8_t irqa_status;
    uint8_t irqa_mask;

    /* IRQ B registers */
    uint8_t irqb_status;
    uint8_t irqb_mask;

    /* FIQ registers */
    uint8_t fiq_status;
    uint8_t fiq_mask;

    /* Timer registers */
    uint16_t timer0_count_latch;
    uint16_t timer0_input_latch;
    uint16_t timer0_output_latch;
    uint16_t timer1_count_latch;
    uint16_t timer1_input_latch;
    uint16_t timer1_output_latch;
    QEMUTimer *timer;
    int64_t timer0_loaded;
    int64_t timer1_loaded;
    int64_t timer0_expire;
    int64_t timer1_expire;

    /* IRQ C registers */
    uint8_t irqc_status;
    uint8_t irqc_mask;

    /* IRQ D registers */
    uint8_t irqd_status;
    uint8_t irqd_mask;

    /* ROM Control registers */
    uint8_t rom_control_0;
    uint8_t rom_control_1;

    /* Mouse registers */
    uint8_t mouse_control;

    /* Sound DMA 0 registers */
    uint32_t snd0_cura;
    uint32_t snd0_enda;
    uint32_t snd0_curb;
    uint32_t snd0_endb;
    uint8_t  snd0_control;
    uint8_t  snd0_status;

    /* Cursor DMA registers */
    uint32_t curs_init;

    /* Video DMA registers */
    uint32_t vidcura;
    uint32_t vidend;
    uint32_t vidstart;
    uint32_t vidinita;
    uint8_t  vidcr;

    /* DMA Interrupt registers */
    uint8_t  dmairq_status;
    uint8_t  dmairq_mask;

    void *vidc_opaque;
    void *ps2_kbd_opaque;
    void *ps2_mouse_opaque;

    struct I2CStruct i2c;
} arm_state;

static uint16_t get_timer0_count(void *opaque);
static uint16_t get_timer1_count_latch(void *opaque);
static void arm_timer_update(void *opaque);

#define KERNEL_ARGS_ADDR 0x100
// 0x8000 + phys_ram_base = 0x10008000 emulated address
#define KERNEL_LOAD_ADDR 0x00008000
#define INITRD_LOAD_ADDR 0x00800000

/* Memory map */
#define RISCPC_PHYS_ROM0    0x00000000
#define RISCPC_PHYS_VRAM    0x02000000
#define RISCPC_PHYS_IO_VIDC 0x03400000
#define RISCPC_PHYS_RAM     0x10000000

#define ARM_TIMER_FREQ 2000000

#define ARM_IRQA_FLYBACK  ((0 << 3) | 3)
#define ARM_IRQA_POWER_ON ((0 << 3) | 4) /* Power-on Reset */
#define ARM_IRQA_TIMER0   ((0 << 3) | 5)
#define ARM_IRQA_TIMER1   ((0 << 3) | 6)

#define ARM_IRQB_KBD_TX   ((1 << 3) | 6)
#define ARM_IRQB_KBD_RX   ((1 << 3) | 7)

#define ARM_IRQD_MOUSE_RX ((3 << 3) | 0)
#define ARM_IRQD_MOUSE_TX ((3 << 3) | 1)

#define IO_CONTROL_FLYBACK (1 << 7)

static void iomd_pic_update(arm_state *s)
{
    int flags;

    assert(s);

#ifdef DEBUG_IO
    {
        uint8_t irqa   = s->irqa_status   & s->irqa_mask;
        uint8_t irqb   = s->irqb_status   & s->irqb_mask;
        uint8_t irqc   = s->irqc_status   & s->irqc_mask;
        uint8_t irqd   = s->irqd_status   & s->irqd_mask;
        uint8_t dmairq = s->dmairq_status & s->dmairq_mask;
        uint8_t fiq    = s->fiq_status    & s->fiq_mask;
        if (irqa || irqb || irqc || irqd || dmairq || fiq) {
            DPRINTF("iomd_pic_update(): Interrupt requested: "
                    "A:%02x B:%02x (C:%02x D:%02x) DMA:%02x F:%02x\n"
                    "iomd_pic_update(): cur_mode=0x%x I=%c F=%c r15=0x%08x\n",
                    irqa, irqb, irqc, irqd, dmairq, fiq,
                    (s->env->uncached_cpsr & CPSR_M),
                    (s->env->uncached_cpsr & CPSR_I) ? '1' : '0',
                    (s->env->uncached_cpsr & CPSR_F) ? '1' : '0',
                    s->env->regs[15]);
        }
    }
#endif

    if (s->parent_irq != -1) {
        flags = (s->irqa_status & s->irqa_mask) || (s->irqb_status & s->irqb_mask) ||
                (s->irqc_status & s->irqc_mask) || (s->irqd_status & s->irqd_mask) ||
                (s->dmairq_status & s->dmairq_mask);
        pic_set_irq_new(s->parent, s->parent_irq, flags != 0);
    }
    if (s->parent_fiq != -1) {
        flags = s->fiq_status & s->fiq_mask;
        pic_set_irq_new(s->parent, s->parent_fiq, flags != 0);
    }
}

static void iomd_pic_set_irq(void *opaque, int irq, int level)
{
    arm_state *s = opaque;
    uint8_t *status = NULL;

    assert(s);

    switch (irq >> 3) {
        case 0: /* IRQ A */  status = &s->irqa_status; break;
        case 1: /* IRQ B */  status = &s->irqb_status; break;
        case 2: /* IRQ C */  status = &s->irqc_status; break;
        case 3: /* IRQ D */  status = &s->irqd_status; break;
        case 4: /* FIQ */    status = &s->fiq_status; break;
        case 5: /* DMAIRQ */ status = &s->dmairq_status; break;
        default: cpu_abort(cpu_single_env, "iomd_pic_set_irq");
    }
    if (level) {
        *status |= 1 << (irq & 7);
    } else {
        *status &= ~(1 << (irq & 7));
    }
    iomd_pic_update(s);
}

static inline uint8_t odd_parity(uint8_t byte)
{
    uint8_t even_parity = ((byte >> 7) & 1) ^ ((byte >> 6) & 1) ^ ((byte >> 5) & 1) ^
    ((byte >> 4) & 1) ^ ((byte >> 3) & 1) ^ ((byte >> 2) & 1) ^
    ((byte >> 1) & 1) ^ (byte & 1);
    return ~even_parity & 1;
}

static void riscpc_io_write(void *opaque, target_phys_addr_t addr,
                          uint32_t value)
{
    arm_state *s = opaque;

    assert(opaque);

    if ((addr & 0xF0) != 0x40) {
        DPRINTF("riscpc_io_write: addr=0x%08x value=0x%08x\n",
                (unsigned) addr, value);
    }

    switch (addr & 0xffff) {
        case 0x0: /* IOCR - I/O Control */
            s->io_control = value;
            I2C_Update(&s->i2c, value & 3);
            break;

        /* Keyboard registers */
        case 0x04: /* KBDDAT */
            DPRINTF("\tkeyboard data write: 0x%02x\n", value & 0xff);
#if 0
// DC code commented out to prevent interupts
            ps2_write_keyboard(s->ps2_kbd_opaque, value & 0xFF);
            /* Trigger keyboard Tx irq */
            // s->irqb_status &= ~ARM_IRQB_KBD_TX;
            // arm_update_irq(opaque);
            // s->irqb_status |= ARM_IRQB_KBD_TX;
            //arm_update_irq(opaque);
#endif
            break;
        case 0x08: /* KBDCR */
            s->kbd_control |= 0x8;
            ///TODO implement this properly
            break;

        /* IRQ A registers */
        case 0x14: /* IRQRQA */
            /* reset appropriate bits in status register */
            s->irqa_status &= ~(value & 0xFF);
            DPRINTF("\tirqa_request write: 0x%02x\n", value & 0xff);
            iomd_pic_update(s);
            break;
        case 0x18: /* IRQMSKA */
            DPRINTF("\tirqa_mask write: 0x%02x\n", value & 0xff);
            s->irqa_mask = value & 0xFF;
            iomd_pic_update(s);
            break;

        /* IRQ B registers */
        case 0x28: /* IRQMSKB */
            DPRINTF("\tirqb_mask write: 0x%02x\n", value & 0xff);
            s->irqb_mask = value & 0xFF;
            iomd_pic_update(s);
            break;

        /* FIQ registers */
        case 0x38:
            DPRINTF("\tfiq_mask write: 0x%02x\n", value & 0xff);
            s->fiq_mask = value & 0xff;
            iomd_pic_update(s);
            break;

        /* Timer 0 registers */
        case 0x40: /* T0LOW */
            s->timer0_input_latch = (s->timer0_input_latch & 0xFF00) | (value & 0xFF);
            break;
        case 0x44: /* T0HIGH */
            s->timer0_input_latch = (s->timer0_input_latch & 0x00FF) | ((value & 0xFF) << 8);
            break;
        case 0x48: /* T0GO */
            arm_timer_update(opaque);
            break;
        case 0x4C: /* T0LAT */
            s->timer0_output_latch = get_timer0_count(opaque);
            break;

        /* Timer 1 registers */
        case 0x50: /* T1LOW */
            s->timer1_input_latch = (s->timer1_input_latch & 0xFF00) | (value & 0xFF);
            break;
        case 0x54: /* T1HIGH */
            s->timer1_input_latch = (s->timer1_input_latch & 0x00FF) | ((value & 0xFF) << 8);
            break;
        case 0x58: /* T1GO */
            arm_timer_update(opaque);
            break;
        case 0x5C: /* T1LAT */
            s->timer1_output_latch = get_timer1_count_latch(opaque);
            break;

        /* IRQ C registers */
        case 0x68: /* IRQMSKC */
            s->irqc_mask = value & 0xFF;
            iomd_pic_update(s);
            break;

        /* IRQ D registers */
        case 0x78: /* IRQMSKD */
            s->irqd_mask = value & 0xFF;
            iomd_pic_update(s);
            break;

        /* ROM Control registers */
        case 0x80:
            s->rom_control_0 = value & 0xFF;
            break;
        case 0x84:
            s->rom_control_1 = value & 0xFF;
            break;

        /* Mouse registers */
        case 0xA4: /* MOUSEY */
            DPRINTF("\tmouse Y write: 0x%02x\n", value & 0xFF);
            break;
        case 0xA8: /* MSECR */
            s->mouse_control |= 0x8;
            ///TODO implement this properly
            break;

        /* Sound DMA 0 registers */
        case 0x180: /* SD0CURA */
            s->snd0_cura = value;
            break;
        case 0x184: /* SD0ENDA */
            s->snd0_enda = value;
            break;
        case 0x188: /* SD0CURB */
            s->snd0_curb = value;
            break;
        case 0x18c: /* SD0ENDB */
            s->snd0_endb = value;
            break;
        case 0x190: /* SD0CR */
            s->snd0_control = value;
            break;

        /* Cursor DMA registers */
        case 0x1c4: /* CURSINIT - Cursor DMA Init */
            s->curs_init = value & 0x1ffffff0;
            break;

        /* Video DMA registers */
        case 0x1D0: /* VIDCURA */
            s->vidcura  = value & 0x1FFFFFF0;
            break;
        case 0x1D4: /* VIDEND */
            s->vidend   = value & 0x1FFFFFF0;
            vidc_update_dma(s->vidc_opaque, s->vidstart, s->vidend,
                            s->vidinita);
            break;
        case 0x1D8: /* VIDSTART */
            s->vidstart = value & 0x1FFFFFF0;
            vidc_update_dma(s->vidc_opaque, s->vidstart, s->vidend,
                            s->vidinita);
            break;
        case 0x1DC: /* VIDINITA */
            // DC s->vidinita = value & 0x1FFFFFF0; DC
            s->vidinita = value;
            vidc_update_dma(s->vidc_opaque, s->vidstart, s->vidend,
                            s->vidinita);
            break;
        case 0x1E0: /* VIDCR */
            s->vidcr = value;
            break;

        /* DMA Interrupt registers */
        case 0x1f4: /* DMARQ - DMA interrupt request */
            /* Ignore writes */
            break;
        case 0x1f8: /* DMAMSK - DMA interrupt mask */
            s->dmairq_mask = value & 0xFF;
            break;

        case 0x0c: /* I/O Port Control - according to RPCEmu */
        case 0x3c: /* Clock Control */
        case 0x6c: /* Video LCD and Serial Sound MUX Control */
        case 0x88: /* ???? refresh period */
        case 0x8c: /* DRAM refresh period */
        case 0x90: /* FLYBACK - flyback line size */
        case 0xc4: /* I/O Timing Control */
        case 0xc8: /* I/O Expansion Card Timing Control */
        case 0xd0: /* DRAMCR DRAM Control */
            /* Nothing to do */
            break;

        default:
            DPRINTF("riscpc_io_write(): unknown addr=0x%08x value=0x%08x\n",
                    (unsigned) addr, value);
            break;
    }
}

static uint32_t riscpc_io_read(void *opaque, target_phys_addr_t addr)
{
    arm_state *s = opaque;
    uint8_t byte;

    assert(opaque);

    if ((addr & 0xf0) != 0x40) {
        DPRINTF("riscpc_io_read: addr=0x%08x\n", (unsigned) addr);
    }

    switch (addr & 0xffff) {
        case 0x0: /* IOCR - I/O Control */
            //DPRINTF("\ti2c read: 0x%x\n", s->i2c.i2c_input | s->io_control);
            return (vidc_get_flyback(s->vidc_opaque) ? IO_CONTROL_FLYBACK : 0) |
                   (s->io_control & 0x7c) | 0x34 | s->i2c.i2c_input;

        /* Keyboard registers */
        case 0x04: /* KBDDAT */
#if 0
// DC code commented out to prevent interupts going off
            byte = ps2_read_data(s->ps2_kbd_opaque);
            s->kbd_control = (s->kbd_control & ~0x20) | (odd_parity(byte) << 2);
            DPRINTF("\tkeyboard data read: 0x%02x (r15=0x%08x)\n",
                    byte, s->env->regs[15]);
            s->irqb_status &= ~ARM_IRQB_KBD_RX;
            arm_update_irq(s);
            return byte;
#endif
            return 0;
        case 0x08: /* KBDCR */
            DPRINTF("\tkeyboard control read: 0x%02x\n", s->kbd_control);
            return s->kbd_control;

        case 0x0c: /* IOLINES - I/O Port Control - according to RPCEmu */
            return 0;

        /* IRQ A registers */
        case 0x10: /* IRQSTA */
            DPRINTF("\tirqa_status read: 0x%02x\n", s->irqa_status | (1 << 7));
            return s->irqa_status | (1 << 7); /* Bit 7 always set when read */
        case 0x14: /* IRQRQA */
            DPRINTF("\tirqa_request read: 0x%02x\n", s->irqa_status & s->irqa_mask);
            return s->irqa_status & s->irqa_mask;
        case 0x18: /* IRQMSKA */
            DPRINTF("\tirqa_mask read: 0x%02x\n", s->irqa_mask);
            return s->irqa_mask;

        /* IRQ B registers */
        case 0x20: /* IRQSTB */
            return s->irqb_status;
        case 0x24: /* IRQRQB */
            return s->irqb_status & s->irqb_mask;
        case 0x28: /* IRQMSKB */
            return s->irqb_mask;

        /* FIQ registers */
        case 0x30: /* FIQST */
            return s->fiq_status;
        case 0x34: /* FIQRQ */
            return s->fiq_status & s->fiq_mask;
        case 0x38: /* FIQMSK */
            return s->fiq_mask;

        /* Timer 0 registers */
        case 0x40: /* T0LOW */
            return s->timer0_output_latch & 0xFF;
        case 0x44: /* T0HIGH */
            return (s->timer0_output_latch & 0xFF00) >> 8;

        /* Timer 1 registers */
        case 0x50: /* T1LOW */
            return s->timer1_output_latch & 0xFF;
        case 0x54: /* T1HIGH */
            return (s->timer1_output_latch & 0xFF00) >> 8;

        /* IRQ C registers */
        case 0x60: /* IRQSTC */
            return s->irqc_status;
        case 0x64: /* IRQRQC */
            return s->irqc_status & s->irqc_mask;
        case 0x68: /* IRQMSKC */
            return s->irqc_mask;

        /* IRQ D registers */
        case 0x70: /* IRQSTD */
            return s->irqd_status | 0x18;
        case 0x74: /* IRQRQD */
            return (s->irqd_status | 0x18) & s->irqd_mask;
        case 0x78: /* IRQMSKD */
            return s->irqd_mask;

        /* ROM Control registers */
        case 0x80: /* ROM control 0 */
            return s->rom_control_0;
        case 0x84: /* ROM control 1 */
            return s->rom_control_1;

        /* Chip ID registers */
        case 0x94: /* ID byte */
            return (CHIP_ID & 0xff);
        case 0x98: /* ID byte */
            return (CHIP_ID >> 8);
        case 0x9C: /* Mysterious chip version byte */
            return 0x0;

        /* Mouse registers */
        case 0xA8: /* MSEDAT */
            byte = ps2_read_data(s->ps2_mouse_opaque);
            s->mouse_control = (s->mouse_control & ~0x20) | (odd_parity(byte) << 2);
            DPRINTF("\tkeyboard data read: 0x%02x (r15=0x%08x)\n",
                    byte, s->env->regs[15]);
            s->irqd_status &= ~ARM_IRQD_MOUSE_RX;
            iomd_pic_update(s);
            return byte;
        case 0xAC: /* MSECR */
            DPRINTF("\tmouse control read: 0x%02x\n", s->mouse_control);
            return s->mouse_control;
            ///TODO implement this properly

        /* Sound DMA 0 registers */
        case 0x194: /* SD0ST Sound DMA 0 Status */
            return s->snd0_status;

        /* Video DMA regsiters */
        case 0x1d0: /* VIDCURA */
            return s->vidcura;
        case 0x1d4: /* VIDEND */
            return s->vidend;
        case 0x1d8: /* VIDSTART */
            return s->vidstart;
        case 0x1dc: /* VIDINITA */
            return s->vidinita;
        case 0x1e0: /* VIDCR */
            return s->vidcr;

        /* DMA Interrupt registers */
        case 0x1f0: /* DMAST - DMA interrupt status */
            return s->dmairq_status;
        case 0x1f4: /* DMARQ - DMA interrupt request */
            return s->dmairq_status & s->dmairq_mask;
        case 0x1f8: /* DMAMSK - DMA interrupt mask*/
            return s->dmairq_mask;

        default:
            DPRINTF("riscpc_io_read(): unknown addr=0x%08x\n",
                    (unsigned) addr);
            return 0;
    }
    return 0;
}

static void riscpc_unknown_write(void *opaque, target_phys_addr_t addr,
                                 uint32_t value)
{
    DPRINTF("%s(): addr=0x%08x value=0x%08x\n", __func__,
            (unsigned) addr, value);
}

static uint32_t riscpc_unknown_read(void *opaque, target_phys_addr_t addr)
{
    ///RPCEMU from there, what does it do?
    if (addr == 0x3310000) {
        DPRINTF("read mouse buttons and sync\n");
        return 0x70;
    }

    DPRINTF("%s(): addr=0x%08x\n", __func__, (unsigned) addr);

    return 0xffffffff;
}


static CPUReadMemoryFunc *riscpc_io_readfn[] = {
    riscpc_io_read,
    riscpc_io_read,
    riscpc_io_read
};

static CPUWriteMemoryFunc *riscpc_io_writefn[] = {
    riscpc_io_write,
    riscpc_io_write,
    riscpc_io_write
};

static CPUReadMemoryFunc *riscpc_unknown_readfn[] = {
    riscpc_unknown_read,
    riscpc_unknown_read,
    riscpc_unknown_read
};

static CPUWriteMemoryFunc *riscpc_unknown_writefn[] = {
    riscpc_unknown_write,
    riscpc_unknown_write,
    riscpc_unknown_write
};

static inline void arm_trigger_irqa_timer0(void *opaque)
{
    assert(opaque);

    pic_set_irq_new(opaque, ARM_IRQA_TIMER0, 1);
}

static inline void arm_trigger_irqa_timer1(void *opaque)
{
    assert(opaque);

    pic_set_irq_new(opaque, ARM_IRQA_TIMER1, 1);
}

static uint16_t get_timer0_count(void *opaque)
{
    arm_state *s = opaque;

    assert(opaque);

    uint32_t diff = muldiv64(qemu_get_clock(vm_clock) - s->timer0_loaded, ARM_TIMER_FREQ, ticks_per_sec);
    return s->timer0_count_latch - (diff % s->timer0_count_latch);
}

static uint16_t get_timer1_count_latch(void *opaque)
{
     //arm_state *s = opaque;

     assert(opaque);

     return -1;
}



static inline void arm_trigger_irqb_ps2_kbd_rx(void *opaque, int level)
{
    arm_state *s = opaque;

    assert(opaque);

    DPRINTF("ps2 kbd trigger: level=%d\n", level);
    s->kbd_control |= 0x20;
    pic_set_irq_new(opaque, ARM_IRQB_KBD_RX, level);
}

static inline void arm_trigger_irqd_ps2_mouse_rx(void *opaque, int level)
{
    arm_state *s = opaque;

    assert(opaque);

    DPRINTF("ps2 mouse trigger: level=%d\n", level);
    s->mouse_control |= 0x20;
    pic_set_irq_new(opaque, ARM_IRQD_MOUSE_RX, level);
}


static void arm_timer_reload_timer0(void *opaque)
{
    int64_t delay;
    arm_state *s = opaque;

    assert(opaque);

    s->timer0_count_latch = s->timer0_input_latch;
    s->timer0_loaded = s->timer0_expire;
    delay = muldiv64(s->timer0_count_latch, ticks_per_sec, ARM_TIMER_FREQ);
    s->timer0_expire += delay;
}

static void arm_timer_update(void *opaque)
{
    int64_t now = qemu_get_clock(vm_clock);
    arm_state *s = opaque;

    assert(opaque);

    if( now - s->timer0_expire >= 0 ) {
        arm_timer_reload_timer0(opaque);
        arm_trigger_irqa_timer0(opaque);
    }

    qemu_mod_timer(s->timer, s->timer0_expire);
}

static void arm_timer_callback(void *opaque)
{
    assert(opaque);

    arm_timer_update(opaque);
    // Just for fun
    //arm_trigger_irqa_flyback(opaque);
}

static void *arm_init(CPUState *env, DisplayState *ds)
{
    int io_regs, io_unknown;
    arm_state *s;
    void *cpu_pic;

    assert(env && ds);

    s = qemu_mallocz(sizeof(arm_state));
    s->env = env;

    /* Initialise interrupt controller */
    cpu_pic = arm_pic_init_cpu(env);
    s->handler = iomd_pic_set_irq;
    s->parent = cpu_pic;
    s->parent_irq = ARM_PIC_CPU_IRQ;
    s->parent_fiq = ARM_PIC_CPU_FIQ;

    io_regs = cpu_register_io_memory(0, riscpc_io_readfn, riscpc_io_writefn, s);
    io_unknown = cpu_register_io_memory(0, riscpc_unknown_readfn,
                                        riscpc_unknown_writefn, s);

    /* Map regions of IO space */
    cpu_register_physical_memory(0x03000000, 0x00200000, io_unknown);
    cpu_register_physical_memory(0x03200000, 0x00010000, io_regs);
    cpu_register_physical_memory(0x03210000, 0x001f0000, io_unknown);
                              /* 0x03400000, 0x00100000  VIDC */
    cpu_register_physical_memory(0x03500000, 0x00b00000, io_unknown);
    cpu_register_physical_memory(0x08000000, 0x08000000, io_unknown);
    /* ARM seems to want to read from up here even though it's not allocated */
    cpu_register_physical_memory(0x80000000, 0x80000000, io_unknown);

    smc37c665_mm_init(0x03010000, 2);

    I2C_Init(&s->i2c);
    s->timer = qemu_new_timer(vm_clock, arm_timer_callback, s);

    s->timer1_count_latch = 0xFFFF;
    s->timer1_input_latch = 0xFFFF;

    s->timer0_count_latch = 0xFFFF;
    s->timer0_input_latch = 0xFFFF;

//    s->ps2_kbd_opaque = ps2_kbd_init( arm_trigger_irqb_ps2_kbd_rx, s);
//    s->ps2_mouse_opaque = ps2_mouse_init( arm_trigger_irqd_ps2_mouse_rx, s);

    pic_set_irq_new(s, ARM_IRQA_POWER_ON, 1);

    s->snd0_status = 0x06; // From RPCEMU

    s->kbd_control = 0xA8;
    s->mouse_control = 0xA8;

    return s;
}

/* The worlds second smallest bootloader.  Set r0-r2, then jump to kernel.  */
static uint32_t bootloader[] = {
  0xe3a00000, /* mov     r0, #0 */
  0xe3a01001, /* mov     r1, #0x1 */ // R1 = 1 => Risc PC
  0x0, /*nop*/ /* WAS orr     r1, r1, #0x100 */
  0xe59f2000, /* ldr     r2, [pc, #0] */
  0xe59ff000, /* ldr     pc, [pc, #0] */
  0, /* Address of kernel args.  Set by integratorcp_init.  */
  0  /* Kernel entry point.  Set by integratorcp_init.  */
};

static void set_kernel_args(uint32_t ram_size, uint32_t vram_size,
                            int initrd_size, const char *kernel_cmdline)
{
    uint32_t *p;

    p = (uint32_t *)(phys_ram_base + KERNEL_ARGS_ADDR);
    /* ATAG_CORE */
    *(p++) = 5;
    *(p++) = 0x54410001;
    *(p++) = 1;
    *(p++) = 0x1000;
    *(p++) = 0;
    /* ATAG_MEM */
    *(p++) = 4;
    *(p++) = 0x54410002;
    *(p++) = ram_size;
    ///TODO setup dram banks properly
    *(p++) = RISCPC_PHYS_RAM;

    if (initrd_size) {
        /* ATAG_INITRD2 */
        *(p++) = 4;
        *(p++) = 0x54420005;
        *(p++) = INITRD_LOAD_ADDR;
        *(p++) = initrd_size;
    }
    if (kernel_cmdline && *kernel_cmdline) {
        /* ATAG_CMDLINE */
        int cmdline_size;

        cmdline_size = strlen(kernel_cmdline);
        memcpy (p + 2, kernel_cmdline, cmdline_size + 1);
        cmdline_size = (cmdline_size >> 2) + 1;
        *(p++) = cmdline_size + 2;
        *(p++) = 0x54410009;
        p += cmdline_size;
    }

#define ATAG_ACORN 0x41000101
    /* Acorn specific information */
    /* struct tag_acorn {
     *     u32 memc_control_reg; // Not used by Risc PC
     *     u32 vram_pages;
     *     u8 sounddefault;
     *     u8 adfsdrives;        // For MFM drives
     *  };
    */
    *(p++) = 5;
    *(p++) = ATAG_ACORN;
    *(p++) = 0;
    *(p++) = vram_size >> 12;
    *(p++) = 0;

    /* ATAG_END */
    *(p++) = 0;
    *(p++) = 0;
}

static void riscpc_init(int ram_size, int vga_ram_size, int boot_device,
                     DisplayState *ds, const char **fd_filename, int snapshot,
                     const char *kernel_filename, const char *kernel_cmdline,
                     const char *initrd_filename, uint32_t cpuid,
                     int vram_size)
{
    const int ram_offset = 0;
    const int vram_offset = ram_offset + ram_size;
    const int rom_offset = vram_offset + vram_size;
    CPUState *env;
    const int linux_boot = (kernel_filename != NULL);
    int kernel_size;
    arm_state *s;

    assert(ds);

    env = cpu_init();
    cpu_arm_set_model(env, cpuid);

    fprintf(stderr,
            "ram_size = %d (%d MiB)\n"
            "vga_ram_size = %d (%d MiB)\n"
            "bios_size = %d (%d MiB)\n",
            ram_size, ram_size >> 20,
            vga_ram_size, vga_ram_size >> 20,
            bios_size, bios_size >> 20);

    /* DRAM banks 0, 1, 2 & 3 */
    cpu_register_physical_memory(RISCPC_PHYS_RAM, ram_size,
                                 ram_offset | IO_MEM_RAM);

    /* VRAM */
    if (vram_size > 0) {
        cpu_register_physical_memory(RISCPC_PHYS_VRAM, vram_size,
                                     vram_offset | IO_MEM_RAM);
    }

    /* ROM bank 0 */
    cpu_register_physical_memory(RISCPC_PHYS_ROM0, 0x00400000,
                                 rom_offset | IO_MEM_ROM);
    /* ROM bank 0 repeated */
    cpu_register_physical_memory(0x20000000, 0x00400000,
                                 rom_offset | IO_MEM_ROM);


    s = arm_init(env, ds);

    /* VIDC */
    s->vidc_opaque = vidc_initialize(ds, RISCPC_PHYS_IO_VIDC,
                                     ram_offset,
                                     RISCPC_PHYS_RAM, ram_size,
                                     vram_offset,
                                     RISCPC_PHYS_VRAM, vram_size,
                                     s, ARM_IRQA_FLYBACK);

    if (linux_boot) {
        printf("Booting ARM Linux\n");

        /* Load the kernel */
        kernel_size = load_image(kernel_filename,
                                 phys_ram_base + KERNEL_LOAD_ADDR);
    } else {
        printf("Booting RISC OS\n");

        kernel_filename = "ROM";

        /* Load the ROM */
        kernel_size = load_image(kernel_filename,
                                 phys_ram_base + rom_offset);
    }

    if (kernel_size < 0) {
        fprintf(stderr, "qemu: could not load kernel '%s'\n", kernel_filename);
        exit(1);
    }

    fprintf(stderr, "kernel_size = %d (%d MiB)\n",
            kernel_size, kernel_size >> 20);

    if (linux_boot) {
        /* Load initial ram disk */
        int initrd_size = 0;

        if (initrd_filename) {
            initrd_size = load_image(initrd_filename,
                                     phys_ram_base + INITRD_LOAD_ADDR);
            if (initrd_size < 0) {
                fprintf(stderr, "qemu: could not load initrd '%s'\n",
                        initrd_filename);
                exit(1);
            }
        }

        /* Prepare bootloader */
        bootloader[5] = KERNEL_ARGS_ADDR + RISCPC_PHYS_RAM;
        bootloader[6] = KERNEL_LOAD_ADDR + RISCPC_PHYS_RAM;
        memcpy(phys_ram_base + rom_offset, bootloader, sizeof(bootloader));
        set_kernel_args(ram_size, vram_size, initrd_size, kernel_cmdline);
    }

    //cpu_set_log(CPU_LOG_TB_OUT_ASM |
    //            CPU_LOG_TB_IN_ASM);
    cpu_set_log(CPU_LOG_INT);
}

static void
riscpc610_init(int ram_size, int vga_ram_size, int boot_device,
               DisplayState *ds, const char **fd_filename, int snapshot,
               const char *kernel_filename, const char *kernel_cmdline,
               const char *initrd_filename)
{
    riscpc_init(ram_size, vga_ram_size, boot_device, ds, fd_filename, snapshot,
                kernel_filename, kernel_cmdline, initrd_filename,
                ARM_CPUID_ARM610, 0);
}

static void
riscpc710_init(int ram_size, int vga_ram_size, int boot_device,
               DisplayState *ds, const char **fd_filename, int snapshot,
               const char *kernel_filename, const char *kernel_cmdline,
               const char *initrd_filename)
{
    riscpc_init(ram_size, vga_ram_size, boot_device, ds, fd_filename, snapshot,
                kernel_filename, kernel_cmdline, initrd_filename,
                ARM_CPUID_ARM710, 2*1024*1024);
}

static void
riscpcsa_init(int ram_size, int vga_ram_size, int boot_device,
              DisplayState *ds, const char **fd_filename, int snapshot,
              const char *kernel_filename, const char *kernel_cmdline,
              const char *initrd_filename)
{
    riscpc_init(ram_size, vga_ram_size, boot_device, ds, fd_filename, snapshot,
                kernel_filename, kernel_cmdline, initrd_filename,
                ARM_CPUID_SA110, 2*1024*1024);
}

QEMUMachine riscpc610_machine = {
    "riscpc610",
    "Acorn Risc PC (ARM610, no VRAM)",
    riscpc610_init,
};

QEMUMachine riscpc710_machine = {
    "riscpc710",
    "Acorn Risc PC (ARM710, 2M VRAM)",
    riscpc710_init,
};

QEMUMachine riscpcsa_machine = {
    "riscpcsa",
    "Acorn Risc PC (StrongARM, 2M VRAM)",
    riscpcsa_init,
};
