/*
 * SMSC 37C665GT Super I/O Chip
 *
 * This code is licenced under the GPL
 */

#include <stdio.h>

#include "vl.h"

#define DEBUG_SMC37C665

#ifdef DEBUG_SMC37C665
# define DPRINTF(fmt, args...) \
  do { fprintf(stderr, "smc37c665: " fmt , ##args); } while (0)
#else
# define DPRINTF(fmt, args...)
#endif

typedef struct {
    uint8_t configregs[16];
    unsigned configmode;
    unsigned configreg;

    target_ulong base;
    unsigned it_shift;
} smc37c665_state;


static uint32_t
smc37c665_ioport_read(void *opaque, uint32_t addr)
{
    smc37c665_state *s = opaque;

    DPRINTF("read addr=0x%03x\n", addr);

    if ((s->configmode == 2) && (addr == 0x3f1)) {
        DPRINTF("read CR%X value=0x%02x\n", s->configreg,
                s->configregs[s->configreg]);
        return s->configregs[s->configreg];
    }

    return 0;
}

static void
smc37c665_ioport_write(void *opaque, uint32_t addr, uint32_t value)
{
    smc37c665_state *s = opaque;

    DPRINTF("write addr=0x%03x value=0x%08x\n", addr, value);

    if (addr == 0x3f0) {
        switch (s->configmode) {
        case 0:
        case 1:
            if (value == 0x55) {
                s->configmode++;
                return;
            }
            break;

        case 2:
            if (value == 0xaa) {
                s->configmode = 0;
                DPRINTF("Cleared config mode\n");
            } else {
                s->configreg = value & 0xf;
                DPRINTF("Selected CR%x\n", s->configreg);
            }
            return;
        }
    } else if (addr == 0x3f1) {
        if (s->configmode == 2) {
            s->configregs[s->configreg] = value;
            DPRINTF("Write CR%0X = 0x%02x\n", s->configreg, value);
            return;
        }
    }

    s->configmode = 0;
}

static uint32_t
smc37c665_mm_readb(void *opaque, target_phys_addr_t addr)
{
    const smc37c665_state *s = opaque;

    return smc37c665_ioport_read(opaque, (addr - s->base) >> s->it_shift)
           & 0xff;
}

static uint32_t
smc37c665_mm_readw(void *opaque, target_phys_addr_t addr)
{
    const smc37c665_state *s = opaque;

    return smc37c665_ioport_read(opaque, (addr - s->base) >> s->it_shift)
           & 0xffff;
}

static uint32_t
smc37c665_mm_readl(void *opaque, target_phys_addr_t addr)
{
    const smc37c665_state *s = opaque;

    return smc37c665_ioport_read(opaque, (addr - s->base) >> s->it_shift);
}

static void
smc37c665_mm_writeb(void *opaque, target_phys_addr_t addr, uint32_t value)
{
    const smc37c665_state *s = opaque;

    smc37c665_ioport_write(opaque, (addr - s->base) >> s->it_shift,
                           value & 0xff);
}

static void
smc37c665_mm_writew(void *opaque, target_phys_addr_t addr, uint32_t value)
{
    const smc37c665_state *s = opaque;

    smc37c665_ioport_write(opaque, (addr - s->base) >> s->it_shift,
                           value & 0xffff);
}

static void
smc37c665_mm_writel(void *opaque, target_phys_addr_t addr, uint32_t value)
{
    const smc37c665_state *s = opaque;

    smc37c665_ioport_write(opaque, (addr - s->base) >> s->it_shift, value);
}

static CPUReadMemoryFunc *
smc37c665_mm_readfn[] = {
    smc37c665_mm_readb,
    smc37c665_mm_readw,
    smc37c665_mm_readl
};

static CPUWriteMemoryFunc *
smc37c665_mm_writefn[] = {
    smc37c665_mm_writeb,
    smc37c665_mm_writew,
    smc37c665_mm_writel
};

void
smc37c665_mm_init(target_ulong base, unsigned it_shift)
{
    smc37c665_state *s;
    int io_mem;

    s = qemu_mallocz(sizeof(smc37c665_state));
    s->configregs[0] = 0x3b;
    s->configregs[1] = 0x9f;
    s->configregs[2] = 0xdc;
    s->configregs[3] = 0x78;
    s->configregs[6] = 0xff;
    s->configregs[13] = 0x65;
    s->configregs[14] = 0x01;
    s->configmode = 0;
    s->base = base;
    s->it_shift = it_shift;

    io_mem = cpu_register_io_memory(0, smc37c665_mm_readfn,
                                       smc37c665_mm_writefn, s);
    cpu_register_physical_memory(base, 0x3ff << it_shift, io_mem);
}
