/*
 * QEMU Acorn VIDC20 Emulator.
 *
 * Copyright (c) 2003 Fabrice Bellard
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#include "vl.h"
#include "vidc.h"

#define DEBUG_VIDC

#ifdef DEBUG_VIDC
# define DPRINTF(fmt, args...) \
  do { fprintf(stderr, fmt, ##args); } while (0)
#else
# define DPRINTF(fmt, args...)
#endif

#define VIDC_FLYBACK_FREQUENCY 60

typedef struct {
    uint8_t r;
    uint8_t g;
    uint8_t b;
} RGB;

typedef unsigned rgb_to_pixel_func(unsigned r, unsigned g, unsigned b);

static void vidc_invalidate_display(void *opaque);
static void vidc_reg_write(void *opaque, target_phys_addr_t /*addr*/, uint32_t val);

typedef struct VIDCState VIDCState;
struct VIDCState {
    /* VIDC registers */
    RGB palette[256];
    unsigned palette_pos;
    RGB border_colour;
    RGB cursor_palette[3];
    uint32_t Horiz_Cycle;
    uint32_t Horiz_SyncWidth;
    uint32_t Horiz_BorderStart;
    uint32_t Horiz_DisplayStart;
    uint32_t Horiz_DisplayEnd;
    uint32_t Horiz_BorderEnd;
    uint32_t Horiz_CursorStart;
    uint32_t Horiz_Interlace;
    uint32_t Vert_Cycle;
    uint32_t Vert_SyncWidth;
    uint32_t Vert_BorderStart;
    uint32_t Vert_DisplayStart;
    uint32_t Vert_DisplayEnd;
    uint32_t Vert_BorderEnd;
    uint32_t Vert_CursorStart;
    uint32_t Vert_CursorEnd;
    uint32_t ControlReg;

    /* Sound registers */
    uint32_t sound_freq;
    uint8_t stereo_image[8];

    /* Display palette */
    uint32_t display_palette[256];
    uint32_t display_border_colour;
    uint32_t display_cursor_palette[3];

    unsigned int bpp;

    /* Redraw flags */
    int invalid;            /* Full display requires redraw */
    int border_invalid;     /* Border requires redraw */
    int cursor_invalid;     /* Cursor requires redraw */

    /* DMA */
    uint32_t vidstart;
    uint32_t vidend;
    uint32_t vidinit;
    uint32_t vidsize;

    /* DRAM & VRAM */
    unsigned long ram_offset;
    uint32_t ram_addr;
    unsigned ram_size;
    unsigned long vram_offset;
    uint32_t vram_addr;
    unsigned vram_size;

    int flyback;

    DisplayState *ds;

   // int64_t timer_loaded;
    int64_t timer_expire;
    int64_t timer_delay;
    QEMUTimer *timer;

    void *pic;
    int irq_flyback;
};

void
vidc_update_dma(void *opaque, uint32_t vidstart, uint32_t vidend,
                uint32_t vidinit)
{
    VIDCState *vs = opaque;

    DPRINTF("%s(): vidstart=0x%08x vidend=0x%08x vidinit=0x%08x\n", __func__,
            vidstart, vidend, vidinit);
    vs->vidstart = vidstart;
    vs->vidend   = vidend;
    vs->vidinit  = vidinit;
    vs->vidsize  = vidend - vidstart;
    vs->invalid = 1;
}

int
vidc_get_flyback(void *opaque)
{
    VIDCState *vs = opaque;
    return vs->flyback;
}

static void vidc_timer_update(VIDCState *vs)
{
    vs->timer_expire += vs->timer_delay;
    qemu_mod_timer(vs->timer, vs->timer_expire);
}

static void vidc_timer_callback(void *opaque)
{
    VIDCState *vs = opaque;

    DPRINTF("Flyback level=%d\n", vs->flyback);
    if (vs->flyback)
        pic_set_irq_new(vs->pic, vs->irq_flyback, 1);
    vs->flyback ^= 1;
    vidc_timer_update(vs);
}

static inline int get_depth_index(const DisplayState *ds)
{
    switch (ds->depth) {
    default:
    case 8:
        return 0;
    case 15:
        return 1;
    case 16:
        return 2;
    case 32:
        if (ds->bgr)
            return 4;
        return 3;
   }
}

static inline unsigned rgb_to_pixel8(unsigned r, unsigned g, unsigned b)
{
    return ((r >> 5) << 5) | ((g >> 5) << 2) | (b >> 6);
}

static inline unsigned rgb_to_pixel15(unsigned r, unsigned g, unsigned b)
{
    return ((r >> 3) << 10) | ((g >> 3) << 5) | (b >> 3);
}

static inline unsigned rgb_to_pixel16(unsigned r, unsigned g, unsigned b)
{
    return ((r >> 3) << 11) | ((g >> 2) << 5) | (b >> 3);
}

static inline unsigned rgb_to_pixel32(unsigned r, unsigned g, unsigned b)
{
    return (r << 16) | (g << 8) | b;
}

static inline unsigned rgb_to_pixel32bgr(unsigned r, unsigned g, unsigned b)
{
    return (b << 16) | (g << 8) | r;
}

static rgb_to_pixel_func *rgb_to_pixel_funcs[] = {
    rgb_to_pixel8,
    rgb_to_pixel15,
    rgb_to_pixel16,
    rgb_to_pixel32,
    rgb_to_pixel32bgr,
};

static void
vidc_update_display_palette(VIDCState *vs, int start, int end)
{
    rgb_to_pixel_func *rgb_to_pixel;
    int i;

    rgb_to_pixel = rgb_to_pixel_funcs[get_depth_index(vs->ds)];

    for (i = start; i < end; i++) {
        vs->display_palette[i] = rgb_to_pixel(vs->palette[i].r,
                                              vs->palette[i].g,
                                              vs->palette[i].b);
    }
    vs->display_border_colour = rgb_to_pixel(vs->border_colour.r,
                                             vs->border_colour.g,
                                             vs->border_colour.b);
    for (i = 0; i < 3; i++) {
        vs->display_cursor_palette[i] = rgb_to_pixel(vs->cursor_palette[i].r,
                                                     vs->cursor_palette[i].g,
                                                     vs->cursor_palette[i].b);
    }
}

static void
vidc_draw_line16(VIDCState *vs, uint8_t *d, int width)
{
    uint16_t *p = (uint16_t *) d;
    int x;

    for (x = 0; x < width; x++) {
        *p++ = vs->display_border_colour;
    }
}

static void
vidc_draw_line32(VIDCState *vs, uint8_t *d, int width)
{
    uint32_t *p = (uint32_t *) d;
    int x;

    for (x = 0; x < width; x++) {
        *p++ = vs->display_border_colour;
    }
}

static void
vidc_update_display(void *opaque)
{
    VIDCState *vs = opaque;
    uint8_t *d;
    void (*f)(VIDCState *vs, uint8_t *d, int width);
    int dd, y;

    if (vs->ds->depth == 0) {
        /* Nothing to do */
        return;
    }

    if (!(vs->border_invalid)) {
        return;
    }

    d = vs->ds->data;
    dd = vs->ds->linesize;

    switch (vs->ds->depth) {
    case 16:
        f = vidc_draw_line16;
        break;
    case 32:
        f = vidc_draw_line32;
        break;

    default:
        fprintf(stderr, "%s(): unsupported depth=%d\n", __func__,
                vs->ds->depth);
        exit(1);
    }

    for (y = 0; y < 16; y++) {
        f(vs, d, vs->ds->width);
        d += dd;
    }

    /* Flush the whole output display to host */
    dpy_update(vs->ds, 0, 0, vs->ds->width, vs->ds->height);

    vs->border_invalid = 0;
}

/* force a full display refresh */
static void vidc_invalidate_display(void *opaque)
{
    VIDCState *vs = opaque;

    vs->invalid = vs->border_invalid = vs->cursor_invalid = 1;
}

static uint32_t vidc_reg_read(void *opaque, target_phys_addr_t addr)
{
    addr &= 0xfffff;
    fprintf(stderr, "vidc_reg_read(): addr=0x%05x\n", (unsigned) addr);

    exit(1);

    return 0;
}

static CPUReadMemoryFunc *vidc_reg_readfn[3] = {
    vidc_reg_read,
    vidc_reg_read,
    vidc_reg_read,
};

static CPUWriteMemoryFunc *vidc_reg_writefn[3] = {
    vidc_reg_write,
    vidc_reg_write,
    vidc_reg_write,
};

static void vidc_save(QEMUFile *f, void *opaque)
{
    //VIDCState *vs = opaque;

    DPRINTF("%s()\n", __func__);
}

static int vidc_load(QEMUFile *f, void *opaque, int version_id)
{
    //VIDCState *vs = opaque;

    DPRINTF("%s()\n", __func__);

    return -EINVAL;
}

/* save the display in a PPM image even if no display is
   available */
static void vidc_screen_dump(void *opaque, const char *filename)
{
}

void *vidc_initialize(DisplayState *ds, uint32_t addr,
                      unsigned long ram_offset,
                      target_phys_addr_t ram_addr, unsigned ram_size,
                      unsigned long vram_offset,
                      target_phys_addr_t vram_addr, unsigned vram_size,
                      void *pic, int irq_flyback)
{
    VIDCState *vs;
    int vidc_io_memory;

    DPRINTF("vidc_initialize(): ram_offset =%luM ram_addr =0x%08x ram_size=%uM\n"
            "                   vram_offset=%luM vram_addr=0x%08x vram_size=%uM\n",
            ram_offset >> 20, (unsigned) ram_addr, ram_size >> 20,
            vram_offset >> 20, (unsigned) vram_addr, vram_size >> 20);

    vs = qemu_mallocz(sizeof(VIDCState));
    if (!vs)
        return NULL;

    vs->ram_offset = ram_offset;
    vs->ram_addr = ram_addr;
    vs->ram_size = ram_size;
    vs->vram_offset = vram_offset;
    vs->vram_addr = vram_addr;
    vs->vram_size = vram_size;

    vs->ds = ds;
    vs->pic = pic;
    vs->irq_flyback = irq_flyback;

    /* Initialise timer for flyback interrupts */
    vs->timer = qemu_new_timer(vm_clock, vidc_timer_callback, vs);
    vs->timer_delay = ticks_per_sec / (VIDC_FLYBACK_FREQUENCY * 2);
    vs->timer_expire = qemu_get_clock(vm_clock);
    vidc_timer_update(vs);

    vidc_io_memory = cpu_register_io_memory(0, vidc_reg_readfn, vidc_reg_writefn, vs);
    cpu_register_physical_memory(addr, 0x100000, vidc_io_memory);

    graphic_console_init(vs->ds, vidc_update_display, vidc_invalidate_display,
                         vidc_screen_dump, vs);

    register_savevm("vidc", 0, 1, vidc_save, vidc_load, vs);

    return vs;
}

static int
vidc_update_palette_entry(RGB *rgb, uint32_t val)
{
    int changed = 0;

    if (rgb->r != (val & 0xff)) {
        rgb->r = val & 0xff;
        changed = 1;
    }
    if (rgb->g != ((val >> 8) & 0xff)) {
        rgb->g = (val >> 8) & 0xff;
        changed = 1;
    }
    if (rgb->b != ((val >> 16) & 0xff)) {
        rgb->b = (val >> 16) & 0xff;
        changed = 1;
    }
    return changed;
}

static void vidc_reg_write(void *opaque, target_phys_addr_t addr, uint32_t val)
{
    VIDCState *vs = opaque;
    int changed;

    DPRINTF("%s(): val=0x%08x (val>>28)=0x%x\n", __func__, val, val >> 28);
    addr &= 0xfffff;
    if (addr != 0x00000) {
        fprintf(stderr, "%s(): unknown address\n", __func__);
        exit(1);
    }

    switch (val >> 28) {
    case 0x0: /* Palette */
        DPRINTF("\tVIDC Palette write 0x%06x\n", val & 0xffffff);
        changed = vidc_update_palette_entry(&vs->palette[vs->palette_pos], val);
        if (changed) {
            vidc_update_display_palette(vs, vs->palette_pos,
                                        vs->palette_pos + 1);
            vs->invalid = 1;
        }
        vs->palette_pos = (vs->palette_pos + 1) & 0xff;
        break;

    case 0x1: /* Palette position */
        DPRINTF("\tVIDC palette pos write val=0x%02x\n", val & 0xff);
        vs->palette_pos = val & 0xff;
        break;

    case 0x4: /* Border col */
        DPRINTF("\tVIDC border colour write val=0x%06x\n", val & 0xffffff);
        changed = vidc_update_palette_entry(&vs->border_colour, val);
        if (changed) {
            vidc_update_display_palette(vs, 0, 0);
            vs->border_invalid = 1;
        }
        break;

    case 0x5:
    case 0x6:
    case 0x7: /* Cursor Palette */
        changed = vidc_update_palette_entry(&vs->cursor_palette[(val >> 28) - 0x5], val);
        if (changed) {
            vidc_update_display_palette(vs, 0, 0);
            vs->cursor_invalid = 1;
        }
        break;

    case 0x8: /* Horizontal */
    case 0x9: /* Vertical */
        switch (val >> 24) {
        case 0x83:
            DPRINTF("\tVIDC Horiz Display Start val=%u\n", val & 0x3ffe);
            vs->Horiz_DisplayStart = val & 0x3ffe;
            break;
        case 0x84:
            DPRINTF("\tVIDC Horiz Display End val=%u\n", val & 0x3ffe);
            vs->Horiz_DisplayEnd = val & 0x3ffe;
            break;

        case 0x93:
            DPRINTF("\tVIDC Vert Display Start val=%u\n", val & 0x1fff);
            vs->Vert_DisplayStart = val & 0x1fff;
            break;
        case 0x94:
            DPRINTF("\tVIDC Vert Display End val=%u\n", val & 0x1fff);
            vs->Vert_DisplayEnd = val & 0x1fff;
            break;
        }
        break;

    case 0xa: /* Stereo Image */
        if (((val >> 24) & 0xf) <= 7) {
            vs->stereo_image[(val >> 24) & 0xf] = val & 0x7;
        }
        break;

    case 0xb: /* Sound Frequency and Control */
        switch ((val >> 24) & 0xf) {
        case 0: /* SFR - Sound Frequency Register */
            break;
        case 1: /* SCTL - Sound Control Register */
            break;
        default:
            DPRINTF("\tVIDC sound register - unknown\n");
            break;
        }
        break;

    case 0xc: /* External */
    case 0xd: /* Frequency Synthesis */
        /* Ignored by RPCEmu */
        break;

    case 0xe:
        {
            int bpp = 0;

            DPRINTF("\tVIDC Control register val=0x%08x\n", val);

            if ((val & 0x0ff00000) != 0) {
                DPRINTF("\tVIDC Control register - some bits should have been zero\n");
            }

            switch ((val >> 5) & 7) {
            case 0: bpp = 1; break;
            case 1: bpp = 2; break;
            case 2: bpp = 4; break;
            case 3: bpp = 8; break;
            case 4: bpp = 16; break;
            case 6: bpp = 32; break;
            }
            vs->bpp = bpp;
            DPRINTF("\tVIDC bpp set to: %u\n", bpp);
            vs->invalid = vs->cursor_invalid = 1;
        }
        break;

    case 0xf: /* DCTL - Data Control Register */
        break;

    default:
        DPRINTF("\tWrite to unknown VIDC register reg=0x%x val=0x%x (val>>28)=0x%x\n",
                (unsigned) addr, val, val >> 28);
        break;
    } /* Register switch */
}
