#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cpu.h"
#include "exec-all.h"

/* Don't attempt StrongArm yet
#define ARM_ARCH_SA
*/

#define CP15_CONTROL_MMU    (1 << 0)
#define CP15_CONTROL_PROG32 (1 << 4)
#define CP15_CONTROL_DATA32 (1 << 5)

void cpu_reset(CPUARMState *env)
{
#if defined (CONFIG_USER_ONLY)
    env->uncached_cpsr = ARM_CPU_MODE_USR;
    env->vfp.xregs[ARM_VFP_FPEXC] = 1 << 30;
#else
    /* SVC mode with interrupts disabled.  */
    /* CPUs that support 26-bit configuration start in mode SVC26.
       Otherwise the start mode is SVC32. */
    if (arm_feature(env, ARM_FEATURE_26BIT_CONFIG)) {
        env->uncached_cpsr = ARM_CPU_MODE_SVC26 | CPSR_A | CPSR_F | CPSR_I;
    } else {
        env->cp15.c1_sys = CP15_CONTROL_PROG32 | CP15_CONTROL_DATA32;
        env->uncached_cpsr = ARM_CPU_MODE_SVC | CPSR_A | CPSR_F | CPSR_I;
    }
    env->vfp.xregs[ARM_VFP_FPEXC] = 0;
#endif
    env->regs[15] = 0;
    env->tlb_flush_countdown = -1;
}

CPUARMState *cpu_arm_init(void)
{
    CPUARMState *env;

    env = qemu_mallocz(sizeof(CPUARMState));
    if (!env)
        return NULL;
    cpu_exec_init(env);
    cpu_reset(env);
    tlb_flush(env, 1);
    return env;
}

static inline void set_feature(CPUARMState *env, int feature)
{
    env->features |= 1u << feature;
}

void cpu_arm_set_model(CPUARMState *env, uint32_t id)
{
    env->cp15.c0_cpuid = id;
    switch (id) {
    case ARM_CPUID_ARM610:
    case ARM_CPUID_ARM710:
        set_feature(env, ARM_FEATURE_26BIT_CONFIG);
        set_feature(env, ARM_FEATURE_PC_STORE_OFFSET_12);
        break;
    case ARM_CPUID_SA110:
        break;
    case ARM_CPUID_ARM926:
        set_feature(env, ARM_FEATURE_VFP);
        env->vfp.xregs[ARM_VFP_FPSID] = 0x41011090;
        break;
    case ARM_CPUID_ARM1026:
        set_feature(env, ARM_FEATURE_VFP);
        set_feature(env, ARM_FEATURE_AUXCR);
        env->vfp.xregs[ARM_VFP_FPSID] = 0x410110a0;
        break;
    default:
        cpu_abort(env, "Bad CPU ID: %x\n", id);
        break;
    }
    cpu_reset(env);
}

void cpu_arm_close(CPUARMState *env)
{
    free(env);
}

#if defined(CONFIG_USER_ONLY)

void do_interrupt (CPUState *env)
{
    env->exception_index = -1;
}

int cpu_arm_handle_mmu_fault (CPUState *env, target_ulong address, int rw,
                              int is_user, int is_softmmu)
{
    if (rw == 2) {
        env->exception_index = EXCP_PREFETCH_ABORT;
        env->cp15.c6_insn = address;
    } else {
        env->exception_index = EXCP_DATA_ABORT;
        env->cp15.c6_data = address;
    }
    return 1;
}

target_ulong cpu_get_phys_page_debug(CPUState *env, target_ulong addr)
{
    return addr;
}

/* These should probably raise undefined insn exceptions.  */
void helper_set_cp15(CPUState *env, uint32_t insn, uint32_t val)
{
    cpu_abort(env, "cp15 insn %08x\n", insn);
}

uint32_t helper_get_cp15(CPUState *env, uint32_t insn)
{
    cpu_abort(env, "cp15 insn %08x\n", insn);
    return 0;
}

void switch_mode(CPUState *env, int mode)
{
    if (mode != ARM_CPU_MODE_USR)
        cpu_abort(env, "Tried to switch out of user mode\n");
}

#else

/* Map CPU modes onto saved register banks.  */
static inline int bank_number (int mode)
{
    /* 26 and 32 bit versions of the same mode
    use the same banks. */
    switch (mode) {
    case ARM_CPU_MODE_USR26:
    case ARM_CPU_MODE_USR:
    case ARM_CPU_MODE_SYS:
        return 0;
    case ARM_CPU_MODE_SVC26:
    case ARM_CPU_MODE_SVC:
        return 1;
    case ARM_CPU_MODE_ABT:
        return 2;
    case ARM_CPU_MODE_UND:
        return 3;
    case ARM_CPU_MODE_IRQ26:
    case ARM_CPU_MODE_IRQ:
        return 4;
    case ARM_CPU_MODE_FIQ26:
    case ARM_CPU_MODE_FIQ:
        return 5;
    }
    cpu_abort(cpu_single_env, "Bad mode %x\n", mode);
    return -1;
}

void switch_mode(CPUState *env, int mode)
{
    int old_mode = env->uncached_cpsr & CPSR_M;
    int old_bank = bank_number(old_mode);
    int new_bank = bank_number(mode);

    fprintf(stderr, "switch_mode: old_mode=0x%x, new_mode=0x%x r15=0x%x\n",
            old_mode, mode, env->regs[15]);

    if (old_bank == new_bank)
        return;

    if (old_bank == 5) {
        memcpy (env->fiq_regs, env->regs + 8, 5 * sizeof(uint32_t));
        memcpy (env->regs + 8, env->usr_regs, 5 * sizeof(uint32_t));
    } else if (new_bank == 5) {
        memcpy (env->usr_regs, env->regs + 8, 5 * sizeof(uint32_t));
        memcpy (env->regs + 8, env->fiq_regs, 5 * sizeof(uint32_t));
    }

    env->banked_r13[old_bank] = env->regs[13];
    env->banked_r14[old_bank] = env->regs[14];
    env->banked_spsr[old_bank] = env->spsr;

    env->regs[13] = env->banked_r13[new_bank];
    env->regs[14] = env->banked_r14[new_bank];
    env->spsr = env->banked_spsr[new_bank];
}

/* Handle a CPU exception.  */
void do_interrupt(CPUARMState *env)
{
    uint32_t addr;
    uint32_t mask;
    int new_mode;
    uint32_t offset;

    fprintf(stderr, "do_interupt: >>>>\n");

    /* If this assert is triggered, we need to implement 26-bit exceptions */
    assert(env->cp15.c1_sys & CP15_CONTROL_PROG32);

    /* TODO: Vectored interrupt controller.  */
    switch (env->exception_index) {
    case EXCP_UDEF:
        fprintf(stderr, "do_interupt: EXCP_UDEF\n");
        new_mode = ARM_CPU_MODE_UND;
        addr = 0x04;
        mask = CPSR_I;
        if (env->thumb)
            offset = 2;
        else
            offset = 4;
        break;
    case EXCP_SWI:
        fprintf(stderr, "do_interupt: EXCP_SWI r0=0x%08x swi=0x%06x\n",
                env->regs[0], ldl_code(env->regs[15] - 4) & 0xffffff);
        new_mode = ARM_CPU_MODE_SVC;
        addr = 0x08;
        mask = CPSR_I;
        /* The PC already points to the next instructon.  */
        offset = 0;
        break;
    case EXCP_PREFETCH_ABORT:
    case EXCP_BKPT:
        fprintf(stderr, "do_interupt: EXCP_PREFETCH_ABORT\n");
        new_mode = ARM_CPU_MODE_ABT;
        addr = 0x0c;
        mask = CPSR_A | CPSR_I;
        offset = 4;
        break;
    case EXCP_DATA_ABORT:
        fprintf(stderr, "do_interupt: EXCP_DATA_ABORT\n");
        new_mode = ARM_CPU_MODE_ABT;
        addr = 0x10;
        mask = CPSR_A | CPSR_I;
        offset = 8;
        break;
    case EXCP_ADDRESS:
        /* Address Exception - 26-bit only */
        new_mode = ARM_CPU_MODE_SVC26;
        addr = 0x14;
        mask = CPSR_I;
        /* TODO - select offset */
        fprintf(stderr, "Address Exception!\n");
        abort();
        break;
    case EXCP_IRQ:
        fprintf(stderr, "do_interupt: EXCP_IRQ\n");
        new_mode = ARM_CPU_MODE_IRQ;
        addr = 0x18;
        /* Disable IRQ and imprecise data aborts.  */
        mask = CPSR_A | CPSR_I;
        offset = 4;
        break;
    case EXCP_FIQ:
        fprintf(stderr, "do_interupt: EXCP_FIQ\n");
        new_mode = ARM_CPU_MODE_FIQ;
        addr = 0x1c;
        /* Disable FIQ, IRQ and imprecise data aborts.  */
        mask = CPSR_A | CPSR_I | CPSR_F;
        offset = 4;
        break;
    default:
        cpu_abort(env, "Unhandled exception 0x%x\n", env->exception_index);
        return; /* Never happens.  Keep compiler happy.  */
    }
    /* High vectors.  */
    if (env->cp15.c1_sys & (1 << 13)) {
        addr += 0xffff0000;
    }
    switch_mode (env, new_mode);
    env->spsr = cpsr_read(env);
    /* Switch to the new mode, and switch to Arm mode.  */
    /* ??? Thumb interrupt handlers not implemented.  */
    fprintf(stderr, "do_interupt: env->regs[15]=0x%x addr=0x%x mode=0x%x\n",env->regs[15],addr, env->uncached_cpsr & 0x1F);
    env->uncached_cpsr = (env->uncached_cpsr & ~CPSR_M) | new_mode;
    env->uncached_cpsr |= mask;
    env->thumb = 0;
    env->regs[14] = env->regs[15] + offset;
    env->regs[15] = addr;
    env->interrupt_request |= CPU_INTERRUPT_EXITTB;

    fprintf(stderr, "do_interupt: <<<<\n");
}

/* Check section/page access permissions.
   Returns the page protection flags, or zero if the access is not
   permitted.  */
static inline int check_ap(CPUState *env, int ap, int domain, int access_type,
                           int is_user)
{
  if (domain == 3)
    return PAGE_READ | PAGE_WRITE;

  switch (ap) {
  case 0:
      if (access_type == 1)
          return 0;
      switch ((env->cp15.c1_sys >> 8) & 3) {
      case 1:
          return is_user ? 0 : PAGE_READ;
      case 2:
          return PAGE_READ;
      default:
          return 0;
      }
  case 1:
      return is_user ? 0 : PAGE_READ | PAGE_WRITE;
  case 2:
      if (is_user)
          return (access_type == 1) ? 0 : PAGE_READ;
      else
          return PAGE_READ | PAGE_WRITE;
  case 3:
      return PAGE_READ | PAGE_WRITE;
  default:
      abort();
  }
}

static int get_phys_addr(CPUState *env, uint32_t address, int access_type,
                         int is_user, uint32_t *phys_ptr, int *prot)
{
    int code;
    uint32_t table;
    uint32_t desc;
    int type;
    int ap;
    int domain;
    uint32_t phys_addr;

    /* Fast Context Switch Extension.  */
    /* Does not exist on ARMv3
    if (address < 0x02000000)
        address += env->cp15.c13_fcse; */

    if ((env->cp15.c1_sys & CP15_CONTROL_MMU) == 0) {
        /* MMU diusabled.  */
        *phys_ptr = address;
        *prot = PAGE_READ | PAGE_WRITE;
    } else {
        /* Pagetable walk.  */
        /* Lookup l1 descriptor.  */
        table = (env->cp15.c2 & 0xffffc000) | ((address >> 18) & 0x3ffc);
        desc = ldl_phys(table);
        type = (desc & 3);
        domain = (env->cp15.c3 >> ((desc >> 4) & 0x1e)) & 3;
        if (type == 0) {
            /* Secton translation fault.  */
            code = 5;
            goto do_fault;
        }
        if (domain == 0 || domain == 2) {
            if (type == 2)
                code = 9; /* Section domain fault.  */
            else
                code = 11; /* Page domain fault.  */
            goto do_fault;
        }
        if (type == 2) {
            /* 1Mb section.  */
            phys_addr = (desc & 0xfff00000) | (address & 0x000fffff);
            ap = (desc >> 10) & 3;
            code = 13;
        } else {
            /* Lookup l2 entry.  */
            table = (desc & 0xfffffc00) | ((address >> 10) & 0x3fc);
            desc = ldl_phys(table);
            switch (desc & 3) {
            case 0: /* Page translation fault.  */
                code = 7;
                goto do_fault;
            case 1: /* 64k page.  */
                phys_addr = (desc & 0xffff0000) | (address & 0xffff);
                ap = (desc >> (4 + ((address >> 13) & 6))) & 3;
                break;
            case 2: /* 4k page.  */
                phys_addr = (desc & 0xfffff000) | (address & 0xfff);
                ap = (desc >> (4 + ((address >> 9) & 6))) & 3;
                break;
            case 3: /* 1k page.  */
                if (type == 1) {
                    /* Page translation fault.  */
                    code = 7;
                    goto do_fault;
                }
                phys_addr = (desc & 0xfffffc00) | (address & 0x3ff);
                ap = (desc >> 4) & 3;
                break;
            default:
                /* Never happens, but compiler isn't smart enough to tell.  */
                abort();
            }
            code = 15;
        }
        *prot = check_ap(env, ap, domain, access_type, is_user);
        if (!*prot) {
    //printf("blah: #4a prot:0%x !*prot:0x%x\n",prot,!*prot);
            /* Access permission fault.  */
            goto do_fault;
        }
        *phys_ptr = phys_addr;
    }
    return 0;
do_fault:
    fprintf(stderr, "mmu fault: 0x%x\n",code | (domain << 4));
    fprintf(stderr, "         : table entry 0x%x\n",
            (env->cp15.c2 & 0xffffc000) | ((address >> 18) & 0x3ffc) );
    fprintf(stderr, "         : address 0x%x\n", address);
    fprintf(stderr, "         : desc 0x%x\n", desc);
    fprintf(stderr, "         : prot 0x%x\n", *prot);
    fprintf(stderr, "         : domain 0x%x\n", domain);
    fprintf(stderr, "         : type 0x%x\n", type);
    fprintf(stderr, "         : phys_addr 0x%x\n", phys_addr);
    int i;
    for(i = 0; i < 16; i++ )
        fprintf(stderr, "reg dump: r%02d=0x%08x\n",i,env->regs[i]);
    return code | (domain << 4);
}

int cpu_arm_handle_mmu_fault (CPUState *env, target_ulong address,
                              int access_type, int is_user, int is_softmmu)
{
    uint32_t phys_addr;
    int prot;
    int ret;

    ret = get_phys_addr(env, address, access_type, is_user, &phys_addr, &prot);
    if (ret == 0) {
        /* Map a single [sub]page.  */
        phys_addr &= ~(uint32_t)0x3ff;
        address &= ~(uint32_t)0x3ff;
        return tlb_set_page (env, address, phys_addr, prot, is_user,
                             is_softmmu);
    }

    if (access_type == 2) {
        /* ARMv3 and ARMv4 do not have separate Fault Status and Fault Address
           for Prefetch Abort. However, updating these here will not cause a
           problem, provided that the Fault Status and Fault Address registers
           only return Data Abort information. */
        env->cp15.c5_insn = ret;
        env->cp15.c6_insn = address;
        env->exception_index = EXCP_PREFETCH_ABORT;
    } else {
        env->cp15.c5_data = ret;
        env->cp15.c6_data = address;
        env->exception_index = EXCP_DATA_ABORT;
    }
    return 1;
}

target_ulong cpu_get_phys_page_debug(CPUState *env, target_ulong addr)
{
    uint32_t phys_addr;
    int prot;
    int ret;

    ret = get_phys_addr(env, addr, 0, 0, &phys_addr, &prot);

    if (ret != 0)
        return -1;

    return phys_addr;
}

void helper_set_cp15(CPUState *env, uint32_t insn, uint32_t val)
{
    uint32_t rm = insn & 0xf;
    uint32_t op2;
    op2 = (insn >> 5) & 7;
    fprintf(stderr, "helper_set_cp15(): reg=%u op2=0x%x val=0x%08x (pc=0x%08x)\n",
            (insn >> 16) & 0xf, op2, val, env->regs[15]);
    switch ((insn >> 16) & 0xf) {
    case 0: /* ID codes.  */
        goto bad_reg;
    case 1: /* System configuration.  */
        switch (op2) {
        case 0:
            fprintf(stderr, "helper_set_cp15(): control write: 0x%08x "
                            "(%c%c%c%c%c%c%c%c%c%c%c%c)\n",
                    val,
                    val & (1 << 11) ? 'Z' : '-',
                    val & (1 << 10) ? 'F' : '-',
                    val & (1 << 9) ? 'R' : '-',
                    val & (1 << 8) ? 'S' : '-',
                    val & (1 << 7) ? 'B' : '-',
                    val & (1 << 6) ? 'L' : '-',
                    val & (1 << 5) ? 'D' : '-',
                    val & (1 << 4) ? 'P' : '-',
                    val & (1 << 3) ? 'W' : '-',
                    val & (1 << 2) ? 'C' : '-',
                    val & (1 << 1) ? 'A' : '-',
                    val & (1 << 0) ? 'M' : '-'
                   );
            env->cp15.c1_sys = val;

            /* If PROG32 not set, and in a 32-bit mode,
               switch to 26-bit version of current mode */
            if (!(val & CP15_CONTROL_PROG32) && (env->uncached_cpsr & 0x10)) {
                switch_mode(env, (env->uncached_cpsr & 0xf));
                env->uncached_cpsr &= ~0x10;
            }

            /* ??? Lots of these bits are not implemented.  */
            /* This may enable/disable the MMU, so do a TLB flush.  */
#ifdef ARM_ARCH_SA
            /* Defer flushing till the next tb (max 3 insns) has been generated */
            env->tlb_flush_countdown = 3;
            fprintf(stderr, "helper_set_cp15: deferring flush\n");
#else
            tlb_flush(env, 1);
#endif
            break;
        case 2:
            env->cp15.c1_coproc = val;
            /* ??? Is this safe when called from within a TB?  */
            tb_flush(env);
        default:
            goto bad_reg;
        }
        break;
    case 2: /* MMU Page table control.  */
        env->cp15.c2 = val;
        /* If Translation Table Base is changed and MMU enabled,
           then the TLB must be flushed */
        if (env->cp15.c1_sys & CP15_CONTROL_MMU) {
            fprintf(stderr, "\tChange of Translation Table Base while MMU enabled\n");
            tlb_flush(env, 1);
        }
        break;
    case 3: /* MMU Domain access control.  */
        env->cp15.c3 = val;
        /* If Domain Access Control is changed and MMU enabled,
           then the TLB must be flushed */
        if (env->cp15.c1_sys & CP15_CONTROL_MMU) {
            fprintf(stderr, "\tChange of Domain Access Control while MMU enabled\n");
            tlb_flush(env, 1);
        }
        break;
    case 4: /* Reserved.  */
        goto bad_reg;
    case 5:
        if (env->cp15.c0_cpuid == ARM_CPUID_ARM610 ||
            env->cp15.c0_cpuid == ARM_CPUID_ARM710)
        {
            /* Flush TLB */
            tlb_flush(env, 1);
            break;
        }
        /* MMU Fault status.  */
        switch (op2) {
        case 0:
            env->cp15.c5_data = val;
            break;
        case 1:
            env->cp15.c5_insn = val;
            break;
        default:
            goto bad_reg;
        }
        break;
    case 6:
        if (env->cp15.c0_cpuid == ARM_CPUID_ARM610 ||
            env->cp15.c0_cpuid == ARM_CPUID_ARM710)
        {
            /* Purge TLB entry */
            tlb_flush(env, 1);
            break;
        }
        /* MMU Fault address.  */
        switch (op2) {
        case 0:
            env->cp15.c6_data = val;
            break;
        case 1:
            env->cp15.c6_insn = val;
            break;
        default:
            goto bad_reg;
        }
        break;
    case 7: /* Cache control.  */
        /* No cache, so nothing to do.  */
        break;
    case 8: /* MMU TLB control.  */
        switch (op2) {
        case 0: /* Invalidate all.  */
            tlb_flush(env, 0);
            break;
        case 1: /* Invalidate single TLB entry.  */
#if 0
            /* ??? This is wrong for large pages and sections.  */
            /* As an ugly hack to make linux work we always flush a 4K
               pages.  */
            val &= 0xfffff000;
            tlb_flush_page(env, val);
            tlb_flush_page(env, val + 0x400);
            tlb_flush_page(env, val + 0x800);
            tlb_flush_page(env, val + 0xc00);
#else
            tlb_flush(env, 1);
#endif
            break;
        default:
            goto bad_reg;
        }
        break;
    case 9: /* Cache lockdown.  */
        switch (op2) {
        case 0:
            env->cp15.c9_data = val;
            break;
        case 1:
            env->cp15.c9_insn = val;
            break;
        default:
            goto bad_reg;
        }
        break;
    case 10: /* MMU TLB lockdown.  */
        /* ??? TLB lockdown not implemented.  */
        break;
    case 11: /* TCM DMA control.  */
    case 12: /* Reserved.  */
        goto bad_reg;
    case 13: /* Process ID.  */
        switch (op2) {
        case 0:
            /* Unlike real hardware the qemu TLB uses virtual addresses,
               not modified virtual addresses, so this causes a TLB flush.
             */
            if (env->cp15.c13_fcse != val)
              tlb_flush(env, 1);
            env->cp15.c13_fcse = val;
            break;
        case 1:
            /* This changes the ASID, so do a TLB flush.  */
            if (env->cp15.c13_context != val)
              tlb_flush(env, 0);
            env->cp15.c13_context = val;
            break;
        default:
            goto bad_reg;
        }
        break;
    case 14: /* Reserved.  */
        goto bad_reg;
    case 15: /* Implementation specific.  */
        if (env->cp15.c0_cpuid == ARM_CPUID_SA110) {
            if ((op2 == 0x2) && (rm == 0x8)) {
                fprintf(stderr, "StrongARM wait for interrupt\n");
                env->exception_index = EXCP_HLT;
                env->halted = 1;
                cpu_loop_exit();
            }
        }
        break;
    }
    return;
bad_reg:
    /* ??? For debugging only.  Should raise illegal instruction exception.  */
    cpu_abort(env, "Unimplemented cp15 register write\n");
}

uint32_t helper_get_cp15(CPUState *env, uint32_t insn)
{
    uint32_t op2;

    op2 = (insn >> 5) & 7;
    fprintf(stderr, "helper_get_cp15(): reg=%u op2=0x%x\n",
            ((insn >> 16) & 0xf), op2);
    switch ((insn >> 16) & 0xf) {
    case 0: /* ID codes.  */
        switch (op2) {
        default: /* Device ID.  */
            return env->cp15.c0_cpuid;
        case 1: /* Cache Type.  */
            return 0x1dd20d2;
        case 2: /* TCM status.  */
            return 0;
        }
    case 1: /* System configuration.  */
        switch (op2) {
        case 0: /* Control register.  */
            return env->cp15.c1_sys;
        case 1: /* Auxiliary control register.  */
            if (arm_feature(env, ARM_FEATURE_AUXCR))
                return 1;
            goto bad_reg;
        case 2: /* Coprocessor access register.  */
            return env->cp15.c1_coproc;
        default:
            goto bad_reg;
        }
    case 2: /* MMU Page table control.  */
        return env->cp15.c2;
    case 3: /* MMU Domain access control.  */
        return env->cp15.c3;
    case 4: /* Reserved.  */
        goto bad_reg;
    case 5: /* MMU Fault status.  */
        switch (op2) {
        case 0:
            return env->cp15.c5_data;
        case 1:
            return env->cp15.c5_insn;
        default:
            goto bad_reg;
        }
    case 6: /* MMU Fault address.  */
        switch (op2) {
        case 0:
            return env->cp15.c6_data;
        case 1:
            /* Arm9 doesn't have an IFAR, but implementing it anyway shouldn't
               do any harm.  */
            return env->cp15.c6_insn;
        default:
            goto bad_reg;
        }
    case 7: /* Cache control.  */
        /* ??? This is for test, clean and invaidate operations that set the
           Z flag.  We can't represent N = Z = 1, so it also clears clears
           the N flag.  Oh well.  */
        env->ZF = 0;
        return 0;
    case 8: /* MMU TLB control.  */
        goto bad_reg;
    case 9: /* Cache lockdown.  */
        switch (op2) {
        case 0:
            return env->cp15.c9_data;
        case 1:
            return env->cp15.c9_insn;
        default:
            goto bad_reg;
        }
    case 10: /* MMU TLB lockdown.  */
        /* ??? TLB lockdown not implemented.  */
        return 0;
    case 11: /* TCM DMA control.  */
    case 12: /* Reserved.  */
        goto bad_reg;
    case 13: /* Process ID.  */
        switch (op2) {
        case 0:
            return env->cp15.c13_fcse;
        case 1:
            return env->cp15.c13_context;
        default:
            goto bad_reg;
        }
    case 14: /* Reserved.  */
        goto bad_reg;
    case 15: /* Implementation specific.  */
        /* ??? Internal registers not implemented.  */
        return 0;
    }
bad_reg:
    /* ??? For debugging only.  Should raise illegal instruction exception.  */
    cpu_abort(env, "Unimplemented cp15 register read\n");
    return 0;
}

#endif

void unaligned(CPUState *env, uint32_t insn, uint32_t addr)
{
    fprintf(stderr, "Unaligned! insn=0x%08x addr=0x%08x\n", insn, addr);

    cpu_dump_state(env, stderr, fprintf, 0);

    exit(1);
}
